use anyhow::{Result, bail};
use assetpath::AssetPath;

use crate::prelude::*;

use std::{collections::HashMap, sync::Arc};

use std::sync::RwLock;

#[derive(Eq, PartialEq, Debug, Clone, Copy)]
#[allow(unused)]
pub enum ControllerType {
    PS4,
    XBox,
    Steam,
    // TODO:
    // Atari
    // Switch
}

#[derive(Clone, Copy, Debug, Default)]

pub struct GraphicsInfo {
    pub sample_count: VkSampleCountFlags,
    pub render_type: SceneType,
    pub render_scale: f32,
    pub vsync: bool,
}

#[derive(Clone, Copy, Debug, Default)]
pub struct RaytracingInfo {
    // software raytracer
    pub triangles_per_as: u32,
    pub accelerate_buffer_size: u32,
    pub stack_size: u32,

    // hardware raytracer
    pub max_lights: u32,
    pub recursion_depth: u32,

    pub use_default_pipeline: bool,

    // shared
    pub max_samplers: u32,
}

#[derive(Clone, Copy, Default, Debug)]
pub struct RasterizerInfo {
    pub shadow_image_size: u32,
    pub enable_lighting: bool,
    pub use_deferred: bool,
    pub use_shadow_maps: bool,
}

#[derive(Debug, Clone, Default)]
pub struct ControllerPictureDirectories {
    pub xbox_path: Option<AssetPath>,
    pub steam_path: Option<AssetPath>,
    pub ps4_path: Option<AssetPath>,
}

#[derive(Debug, Clone, Default)]
pub struct ButtonBackgrounds {
    pub light: AssetPath,
    pub dark: AssetPath,
}

#[derive(Debug, Clone, Default)]
pub struct AssetDirectories {
    pub entity_file_directory: AssetPath,
    pub gltf_file_directory: AssetPath,
    pub tile_file_directory: AssetPath,
}

#[derive(Debug)]
pub struct EngineSettings {
    current_controller_type: RwLock<Option<ControllerType>>,

    controller_buttons: RwLock<HashMap<ControllerButton, Arc<Image>>>,

    graphics_info: RwLock<GraphicsInfo>,
    raytracing_info: RwLock<RaytracingInfo>,
    rasterizer_info: RwLock<RasterizerInfo>,

    controller_directories: ControllerPictureDirectories,
    asset_directories: AssetDirectories,

    pub resource_base_path: String,

    //
    // context: Arc<Context>,
    dark_key: Option<Arc<Image>>,

    light_key: Option<Arc<Image>>,
}

impl EngineSettings {
    #[allow(clippy::too_many_arguments)]
    pub fn new(
        context: &Context,
        mut graphics_info: GraphicsInfo,
        raytracing_info: RaytracingInfo,
        rasterizer_info: RasterizerInfo,
        controller_directories: ControllerPictureDirectories,
        asset_directories: AssetDirectories,
        button_background: ButtonBackgrounds,
        resource_base_path: String,
    ) -> Result<EngineSettings> {
        let dark_key = if button_background.dark.is_empty() {
            None
        } else {
            Some(
                Image::from_file(button_background.dark)?
                    .attach_sampler(Sampler::nearest_sampler().build(context.device())?)
                    .build(context.device(), context.queue())?,
            )
        };

        let light_key = if button_background.light.is_empty() {
            None
        } else {
            Some(
                Image::from_file(button_background.light)?
                    .attach_sampler(Sampler::nearest_sampler().build(context.device())?)
                    .build(context.device(), context.queue())?,
            )
        };

        if graphics_info.render_scale == 0.0 {
            graphics_info.render_scale = 1.0;
        }

        Ok(EngineSettings {
            current_controller_type: RwLock::new(None),

            controller_buttons: RwLock::new(HashMap::new()),

            graphics_info: RwLock::new(graphics_info),
            raytracing_info: RwLock::new(raytracing_info),
            rasterizer_info: RwLock::new(rasterizer_info),

            controller_directories,
            asset_directories,

            resource_base_path,

            dark_key,

            light_key,
        })
    }

    // directories
    pub fn entity_directory(&self) -> &AssetPath {
        &self.asset_directories.entity_file_directory
    }

    pub fn gltf_directory(&self) -> &AssetPath {
        &self.asset_directories.gltf_file_directory
    }

    pub fn tiles_directory(&self) -> &AssetPath {
        &self.asset_directories.tile_file_directory
    }

    pub fn graphics_info(&self) -> Result<GraphicsInfo> {
        Ok(*self.graphics_info.read().unwrap())
    }

    pub fn raytracing_info(&self) -> Result<RaytracingInfo> {
        Ok(*self.raytracing_info.read().unwrap())
    }

    pub fn rasterizer_info(&self) -> Result<RasterizerInfo> {
        Ok(*self.rasterizer_info.read().unwrap())
    }

    pub fn change_scene_type(&self, scene_type: SceneType) -> Result<()> {
        self.graphics_info.write().unwrap().render_type = scene_type;

        Ok(())
    }

    pub fn dark_key(&self) -> &Option<Arc<Image>> {
        &self.dark_key
    }

    pub fn light_key(&self) -> &Option<Arc<Image>> {
        &self.light_key
    }

    pub fn controller_icon(
        &self,
        context: &Context,
        button: ControllerButton,
    ) -> Result<Arc<Image>> {
        Ok(self
            ._controller_icon(context, button)?
            .unwrap_or(self.controller_button_for(context, button, ControllerType::XBox)?))
    }

    fn _controller_icon(
        &self,
        context: &Context,
        button: ControllerButton,
    ) -> Result<Option<Arc<Image>>> {
        // TODO: don't check it every time we request a button
        // self.check_controller(context)?;

        match self.current_controller_type.read().unwrap().as_ref() {
            Some(controller_type) => Ok(Some(self.controller_button_for(
                context,
                button,
                *controller_type,
            )?)),
            None => Ok(None),
        }
    }

    pub fn controller_button_for(
        &self,
        context: &Context,
        button: ControllerButton,
        controller: ControllerType,
    ) -> Result<Arc<Image>> {
        let mut map = self.controller_buttons.write().unwrap();

        if let Some(image) = map.get(&button) {
            return Ok(image.clone());
        }

        let image_path_opt = match controller {
            ControllerType::PS4 => self.controller_directories.ps4_path.as_ref().map(|path| {
                AssetPath::from((path.full_path(), Self::ps4_input_map_to_texture(&button)))
            }),
            ControllerType::Steam => self.controller_directories.steam_path.as_ref().map(|path| {
                AssetPath::from((path.full_path(), Self::steam_input_map_to_texture(&button)))
            }),

            ControllerType::XBox => self.controller_directories.xbox_path.as_ref().map(|path| {
                AssetPath::from((path.full_path(), Self::xbox_input_map_to_texture(&button)))
            }),
        };

        let image_path = match image_path_opt {
            Some(p) => p,
            None => bail!("path not set"),
        };

        let image = Image::from_file(image_path)?
            .attach_sampler(Sampler::nearest_sampler().build(context.device())?)
            .build(context.device(), context.queue())?;

        map.insert(button, image.clone());

        Ok(image)
    }
}

impl EngineSettings {
    fn ps4_input_map_to_texture(button: &ControllerButton) -> &'static str {
        match button {
            ControllerButton::A => "PS4_Cross.png",
            ControllerButton::B => "PS4_Circle.png",
            ControllerButton::X => "PS4_Square.png",
            ControllerButton::Y => "PS4_Triangle.png",
            ControllerButton::Start => "PS4_Options.png",
            ControllerButton::Select => "PS4_Share.png",
            ControllerButton::RightButton => "PS4_R1.png",
            ControllerButton::LeftButton => "PS4_L1.png",
            ControllerButton::RightTrigger => "PS4_R2.png",
            ControllerButton::LeftTrigger => "PS4_L2.png",
            ControllerButton::DPadUp => "PS4_DPad_Up.png",
            ControllerButton::DPadDown => "PS4_DPad_Down.png",
            ControllerButton::DPadRight => "PS4_DPad_Right.png",
            ControllerButton::DPadLeft => "PS4_DPad_Left.png",
            ControllerButton::LeftStick => "PS4_Left_Stick.png",
            ControllerButton::RightStick => "PS4_Right_Stick.png",

            ControllerButton::Misc => todo!("Misc not present on PSController"),
            ControllerButton::Paddle1 => todo!("Paddle1 not present on PSController"),
            ControllerButton::Paddle2 => todo!("Paddle2 not present on PSController"),
            ControllerButton::Paddle3 => todo!("Paddle3 not present on PSController"),
            ControllerButton::Paddle4 => todo!("Paddle4 not present on PSController"),
            ControllerButton::Touchpad => todo!("Touchpad not present on PSController"),

            ControllerButton::Guide => todo!("missing PlayStation icon"),
        }
    }

    fn steam_input_map_to_texture(button: &ControllerButton) -> &'static str {
        match button {
            ControllerButton::A => "Steam_A.png",
            ControllerButton::B => "Steam_B.png",
            ControllerButton::X => "Steam_X.png",
            ControllerButton::Y => "Steam_Y.png",
            ControllerButton::Start => "Steam_Start.png",
            ControllerButton::Select => "Steam_Back.png",
            ControllerButton::RightButton => "Steam_RB.png",
            ControllerButton::LeftButton => "Steam_LB.png",
            ControllerButton::RightTrigger => "Steam_RT.png",
            ControllerButton::LeftTrigger => "Steam_LT.png",
            ControllerButton::DPadUp => "Steam_Left_Track_Up.png",
            ControllerButton::DPadDown => "Steam_Left_Track_Down.png",
            ControllerButton::DPadRight => "Steam_Left_Track_Right.png",
            ControllerButton::DPadLeft => "Steam_Left_Track_Left.png",
            ControllerButton::LeftStick => "Steam_Stick.png",
            ControllerButton::RightStick => "Steam_Right_Track_Center.png",
            ControllerButton::Guide => "Steam_System.png",

            ControllerButton::Misc => todo!("Misc not present on SteamController"),
            ControllerButton::Paddle1 => todo!("Paddle1 not present on SteamController"),
            ControllerButton::Paddle2 => todo!("Paddle2 not present on SteamController"),
            ControllerButton::Paddle3 => todo!("Paddle3 not present on SteamController"),
            ControllerButton::Paddle4 => todo!("Paddle4 not present on SteamController"),
            ControllerButton::Touchpad => todo!("Touchpad not present on SteamController"),
        }
    }

    fn xbox_input_map_to_texture(button: &ControllerButton) -> &'static str {
        match button {
            ControllerButton::A => "360_A.png",
            ControllerButton::B => "360_B.png",
            ControllerButton::X => "360_X.png",
            ControllerButton::Y => "360_Y.png",
            ControllerButton::Start => "360_Start_Alt.png",
            ControllerButton::Select => "360_Back_Alt.png",
            ControllerButton::RightButton => "360_RB.png",
            ControllerButton::LeftButton => "360_LB.png",
            ControllerButton::RightTrigger => "360_RT.png",
            ControllerButton::LeftTrigger => "360_LT.png",
            ControllerButton::DPadUp => "360_Dpad_Up.png",
            ControllerButton::DPadDown => "360_Dpad_Down.png",
            ControllerButton::DPadRight => "360_Dpad_Right.png",
            ControllerButton::DPadLeft => "360_Dpad_Left.png",
            ControllerButton::LeftStick => "360_Left_Stick.png",
            ControllerButton::RightStick => "360_Right_Stick.png",

            ControllerButton::Misc => todo!("Misc not present on XBoxController"),
            ControllerButton::Paddle1 => todo!("Paddle1 not present on XBoxController"),
            ControllerButton::Paddle2 => todo!("Paddle2 not present on XBoxController"),
            ControllerButton::Paddle3 => todo!("Paddle3 not present on XBoxController"),
            ControllerButton::Paddle4 => todo!("Paddle4 not present on XBoxController"),
            ControllerButton::Touchpad => todo!("Touchpad not present on XBoxController"),

            ControllerButton::Guide => todo!("missing XBox icon"),
        }
    }

    // fn check_controller(&self, context: &Context) -> Result<()> {
    //     let active_controller = context.active_controller();

    //     match &active_controller {
    //         // there is an controller connected
    //         Some(controller) => {
    //             let controller_name = controller.read().unwrap().name().clone();

    //             // predict controller type from name
    //             let controller_type = if controller_name.contains("PS4") {
    //                 ControllerType::PS4
    //             } else if controller_name.contains("Steam") {
    //                 ControllerType::Steam
    //             } else {
    //                 ControllerType::XBox
    //             };

    //             let mut current_controller_type = self.current_controller_type.write().unwrap();

    //             // check if current controller type is set

    //             if current_controller_type.is_some() {
    //                 // if controller type is set, check if it is different
    //                 if *current_controller_type.as_ref().unwrap() != controller_type {
    //                     *current_controller_type = Some(controller_type);

    //                     let mut map = self.controller_buttons.write().unwrap();

    //                     // clear map if filled
    //                     if !map.is_empty() {
    //                         map.clear();
    //                     }
    //                 }
    //             } else {
    //                 // set the controller type
    //                 *current_controller_type = Some(controller_type);
    //             }
    //         }
    //         // there is no controller connected
    //         None => {
    //             let mut current_controller_type = self.current_controller_type.write().unwrap();
    //             let mut map = self.controller_buttons.write().unwrap();

    //             // clear current type if set
    //             if current_controller_type.is_some() {
    //                 *current_controller_type = None;
    //             }

    //             // clear map if filled
    //             if !map.is_empty() {
    //                 map.clear();
    //             }
    //         }
    //     }

    //     Ok(())
    // }
}
